from __future__ import division, absolute_import, print_function

__all__ = ['Platform', 'parse_platforms']


class Platform(object):
    all = []

    def __init__(self, name, nickname, sdks, output_directory_name):
        self.name = name
        # TODO: Change to "display name"?
        self.nickname = nickname
        self.sdks = sdks
        self.output_directory_name = output_directory_name

    @classmethod
    def platform_for_nickname(cls, nickname):
        # type: (str) -> Platform
        for platform in cls.all:
            if platform.nickname.lower() == nickname.lower():
                return platform
        return None

    @property
    def device_sdk(self):
        return self.sdks[0]

    def __repr__(self):
        return self.nickname


Platform.all = [
    Platform(name='iOS', nickname='iOS', sdks=['iphoneos', 'iphonesimulator'], output_directory_name='iOS'),
    Platform(name='macOS', nickname='Mac', sdks=['macosx'], output_directory_name='Mac'), # TODO: Change path to macOS? Will SDK name change?
    Platform(name='watchOS', nickname='watchOS', sdks=['watchos', 'watchsimulator'], output_directory_name='watchOS'),
    Platform(name='tvOS', nickname='tvOS', sdks=['appletvos', 'appletvsimulator'], output_directory_name='tvOS')
]


def parse_platforms(s):
    # type: (str) -> [Platform]
    if not s:
        return Platform.all
    else:
        return [Platform.platform_for_nickname(platform.strip()) for platform in s.split(',')]




import click
import networkx as nx
import os
import logging
import sys

from punic.errors import *
from punic.utilities import *
from punic.config import *
from punic.runner import *

def make_graph(punic, open):
    try:
        with error_handling():

            graph = punic.graph()

            logging.info('Writing graph file to "{}".'.format(os.getcwd()))
            nx.drawing.nx_pydot.write_dot(graph, 'graph.dot')

            command = 'dot graph.dot -ograph.png -Tpng'
            if runner.can_run(command):
                logging.info('Rendering dot file to png file.')
                runner.check_run(command)
                if open:
                    click.launch('graph.png')
            else:
                logging.warning('graphviz not installed. Cannot convert graph to a png.')
    except ImportError as e:
        pip = 'pip' if sys.version_info.major < 3 else 'pip3'
        steps = ['brew install graphviz', '{} install pydotplus'.format(pip)]
        lines = ['To enable graph generation please do the following:'] + ['\t' + step for step in steps]
        logging.error('\n'.join(lines))



from __future__ import division, absolute_import, print_function

__all__ = ['Resolver', 'Node']

from collections import (defaultdict, namedtuple)
from networkx import (DiGraph, dfs_preorder_nodes, topological_sort, number_of_nodes, number_of_edges)
import logging

from .repository import *

Node = namedtuple('Node', 'identifier version')


class Resolver(object):
    def __init__(self, root, dependencies_for_node):
        self.root = root
        self.dependencies_for_node = dependencies_for_node

    def build_graph(self, dependency_filter=None):
        # type: ([str]) -> DiGraph

        def populate_graph(graph, parent, depth=0):
            graph.add_node(parent)

            for child_identifier, child_versions in self._dependencies_for_node(parent):
                for child_version in child_versions:
                    child = Node(child_identifier, child_version)
                    if dependency_filter and dependency_filter(child.identifier, child.version) == False:
                        continue
                    graph.add_edge(parent, child)
                    populate_graph(graph, child, depth=depth + 1)

        graph = DiGraph()
        populate_graph(graph=graph, parent=self.root)
        return graph

    def resolve(self):
        # type: () -> DiGraph

        for dependency, revisions in self._dependencies_for_node(self.root):
            logging.debug('<ref>{}</ref> <rev>{}</rev>'.format(dependency, revisions))

        logging.debug('Building universal graph')

        # Build a graph up of _all_ version of _all_ dependencies
        graph = self.build_graph()

        logging.debug('Universal graph has {} nodes, {} edges.'.format(number_of_nodes(graph), number_of_edges(graph)))

        # Build a dictionary of all versions of all dependencies
        all_dependencies = defaultdict(set)
        for dependency, version in dfs_preorder_nodes(graph):
            all_dependencies[dependency].add(version)

        ################################################################################################################

        def prune_1():
            for dependency, versions in all_dependencies.items():
                if len(versions) <= 1:
                    continue
                for version in sorted(versions):
                    if len(graph.predecessors(Node(dependency, version))) <= 1:
                        graph.remove_node(Node(dependency, version))
                        all_dependencies[dependency].remove(version)
                    if len(versions) <= 1:
                        break

        def prune_2():
            def prune(node):
                mini = defaultdict(set)
                if not node in graph:
                    return

                for dependency, version in graph.successors(node):
                    mini[dependency].add(version)
                for dependency, versions in mini.items():
                    some = all_dependencies[dependency]
                    difference = some.difference(versions)
                    for version in difference:
                        graph.remove_node(Node(dependency, version))
                        all_dependencies[dependency].remove(version)
                for successor in graph.successors(node):
                    prune(successor)

            prune(self.root)

        ################################################################################################################

        logging.debug('<sub>Pruning graph</sub>')

        prune_1()
        prune_2()

        logging.debug('Pruned universal graph has {} nodes, {} edges.'.format(number_of_nodes(graph), number_of_edges(graph)))

        ################################################################################################################

        dependencies = set([(dependency, sorted(versions)[-1]) for dependency, versions in sorted(all_dependencies.items())])

        ################################################################################################################

        graph = self.build_graph(dependency_filter=lambda child, child_version: (child, child_version) in dependencies)

        logging.debug('Pruned universal graph has {} nodes, {} edges.'.format(number_of_nodes(graph), number_of_edges(graph)))

        ################################################################################################################

        return graph

    def resolve_build_order(self):
        # type: () -> [(ProjectIdentifier, Revision)]
        graph = self.resolve()
        logging.debug('<sub>Topologically sorting graph</sub>')
        build_order = topological_sort(graph, reverse=True)
        return build_order

    def resolve_versions(self, dependencies):
        # type: (ProjectIdentifier, Revision) -> [ProjectIdentifier, Tag]
        """Given an array of project identifier/version pairs work out the build order"""

        graph = DiGraph()
        versions_for_identifier = dict(dependencies)
        for identifier, version in dependencies:
            parent = Node(identifier, version)
            graph.add_node(parent)

            assert isinstance(version, Revision)

            dependencies_for_node = self._dependencies_for_node(Node(identifier, version))

            for dependency, _ in dependencies_for_node:
                version = versions_for_identifier[dependency]
                child = Node(dependency, version)
                graph.add_edge(parent, child)
        build_order = topological_sort(graph, reverse=True)

        return build_order

    def _dependencies_for_node(self, node):
        # type: (Edge, bool) -> [Any, [Any]]
        return self.dependencies_for_node(node)


def dump(stream, graph, node, depth=0):
    count = len(graph.predecessors(node))

    stream.write("{}{} {}\n".format('\t' * depth, node, count))
    for child in sorted(graph[node].keys()):
        dump(stream, graph, child, depth + 1)



from __future__ import division, absolute_import, print_function

__all__ = ['Config', 'config']

from pathlib2 import Path
import yaml
import logging
import os
import six

from .runner import *
from .xcode import *
from .platform import *


# TODO: This all needs to be cleaned up and made more generic. More configs will be added over time and this will only get worse
# TODO: Allow config file to be relocated and specified on command line
# TODO: Allow subcommands to easily override configs

class Config(object):
    def __init__(self):
        self.xcode = None
        self.repo_overrides = dict()

        self.root_path = Path.cwd()  # type: Path

        self.library_directory = Path('~/Library/Application Support/io.schwa.Punic').expanduser()
        if not self.library_directory.exists():
            self.library_directory.mkdir(parents=True)
        self.repo_cache_directory = self.library_directory / 'repo_cache'
        if not self.repo_cache_directory.exists():
            self.repo_cache_directory.mkdir(parents=True)
        self.punic_path = self.root_path / 'Carthage'
        self.build_path = self.punic_path / 'Build'
        self.checkouts_path = self.punic_path / 'Checkouts'

        self.derived_data_path = self.library_directory / "DerivedData"

        self.platforms = Platform.all
        self.configuration = None

        self.fetch = False
        self.xcode = Xcode.default()

        self.toolchain = None
        self.dry_run = False
        self.use_submodules = False
        self.use_ssh = False

        self.skips = []

        self.verbose = False
        self.echo = False

        self.continuous_integration = 'CI' in os.environ
        if self.continuous_integration:
            logging.info("Running on continuous integration")

        # Read in defaults from punic.yaml (or punic.yml if that exists)
        punic_configuration_path = Path('punic.yaml')
        if not punic_configuration_path.exists():
            punic_configuration_path = Path('punic.yml')
        if punic_configuration_path.exists():
            self.read(punic_configuration_path)
        runner.cache_path = self.library_directory / "cache.shelf"

    def update(self, **kwargs):
        for key, value in sorted(kwargs.items()):
            if value:
                if hasattr(self, key):
                    setattr(self, key, value)

        # Special case for platforms
        platform = kwargs['platform'] if 'platform' in kwargs else None
        if platform:
            self.platforms = parse_platforms(platform)

        if self.verbose and os.environ.get('DUMP_CONFIG', False):
            self.dump()

    def dump(self):

        logging.info('# Environment ##' + '#' * 64)

        logging.info('CWD: {}'.format(os.getcwd()))

        key_width = max([len(k) for k in os.environ.keys()] + [len(k) for k in self.__dict__.items()])

        os.environ.keys()

        for key, value in sorted(os.environ.items()):
            logging.info('{:{key_width}}: {}'.format(key, value, key_width = key_width + 1))

        logging.info('# Configuration ' + '#' * 64)

        for key, value in sorted(self.__dict__.items()):
            logging.info('{:{key_width}}: {}'.format(key, value, key_width = key_width + 1))
        logging.info('#' * 80)

    @property
    def xcode_version(self):
        return self.xcode.version if self.xcode else None

    @xcode_version.setter
    def xcode_version(self, value):
        xcode = Xcode.with_version(value)
        if value and not xcode:
            raise Exception('Could not find xcode version: {}'.format(value))
        if not xcode:
            xcode = Xcode.default()
        self.xcode = xcode

    def read(self, path):
        # type: (Path)

        d = yaml.safe_load(path.open())
        if not d:
            return
        if 'defaults' in d:
            defaults = d['defaults']
            if 'configuration' in defaults:
                self.configuration = defaults['configuration']
            if 'platforms' in defaults:
                self.platforms = parse_platforms(defaults['platforms'])
            elif 'platform' in defaults:
                self.platforms = parse_platforms(defaults['platform'])
            if 'xcode-version' in defaults:
                self.xcode_version = six.text_type(defaults['xcode-version'])

            if 'use-ssh' in defaults:
                self.use_ssh = defaults['use-ssh']

        if 'repo-overrides' in d:
            self.repo_overrides = d['repo-overrides']

        if 'skips' in d:
            self.skips = d['skips'] or []

config = Config()



from __future__ import division, absolute_import, print_function

__all__ = ['Repository', 'Revision']

from flufl.enum import Enum
import functools
import hashlib
import affirm
import six
import logging

from memoize import mproperty
from pathlib2 import Path

# Ideally we could six.urllib but this causes problem with nosetests!
if six.PY2:
    import urlparse
elif six.PY3:
    import urllib.parse as urlparse

from .runner import *
from .config import *
from .errors import *
from .cartfile import *
from .semantic_version import *
from .errors import NoSuchRevision


class Repository(object):
    def __init__(self, punic, identifier, repo_path=None):
        self.punic = punic
        self.identifier = identifier
        if repo_path:
            self.path = repo_path
        else:
            remote_url = self.identifier.remote_url.encode("utf-8")
            url_hash = hashlib.md5(remote_url).hexdigest()
            self.path = punic.config.repo_cache_directory / "{}_{}".format(self.identifier.project_name, url_hash)

        self.specifications_cache = dict()

    def __repr__(self):
        return str(self.identifier)

    def __eq__(self, other):
        return self.identifier == other.identifier

    def __lt__(self, other):
        return self.identifier < other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def check_work_directory(self):
        if not self.path.exists():
            raise RepositoryNotClonedError()

    @property
    def config(self):
        return self.punic.config

    @mproperty
    def tags(self):
        """Return a list of Tag objects representing git tags. Only tags that are valid semantic versions are returned"""
        # type: () -> [Tag]

        self.check_work_directory()

        output = runner.check_run('git tag', cwd=self.path)
        tags = [tag for tag in output.split('\n') if tag]

        if config.verbose == True:
            bad_tags = [tag for tag in tags if not SemanticVersion.is_semantic(tag)]
            if bad_tags:
                logging.warning("<err>Warning</err>: Found tags in \'{}\' that are not semantic: {}".format(self, ', '.join(['\'{}\''.format(tag) for tag in bad_tags])))

        tags = [Revision(repository=self, revision=tag, revision_type=Revision.Type.tag) for tag in tags if SemanticVersion.is_semantic(tag)]
        return sorted(tags)

    def rev_parse(self, s):
        # type: (str) -> str

        self.check_work_directory()

        result = runner.run('git rev-parse "{}"'.format(s), echo=False, cwd=self.path)
        if result.return_code == 0:
            return result.stdout.strip()

        # TODO: assumes remote is called origin.
        command = 'git rev-parse "origin/{}"'.format(s)
        result = runner.run(command, echo=False, cwd=self.path)
        if result.return_code == 0:
            return result.stdout.strip()
        raise Exception('{}: \'{}\' failed with {}'.format(self, command, result))

    def checkout(self, revision):
        # type: (Revision)
        logging.debug('Checking out <ref>{}</ref> @ revision <rev>{}</rev>'.format(self, revision))
        self.check_work_directory()
        try:
            runner.check_run('git checkout "{}"'.format(revision.sha), cwd=self.path)
        except Exception:
            raise NoSuchRevision(repository=self, revision=revision)

        runner.check_run('git submodule update --init --recursive', cwd=self.path)


    def fetch(self):
        if not self.path.exists():
            logging.debug('<sub>Cloning</sub>: <ref>{}</ref>'.format(self))

            url = self.identifier.remote_url

            parsed_url = urlparse.urlparse(url)
            if parsed_url.scheme == 'file':
                repo = parsed_url.path
            else:
                repo = url

            runner.check_run('git clone --recursive --recurse-submodules "{}" "{}"'.format(repo, str(self.path)), cwd=self.path.parent)
        else:
            self.check_work_directory()

            logging.info('<sub>Fetching</sub>: <ref>{}</ref>'.format(self))
            runner.check_run('git fetch', cwd=self.path)


    def specifications_for_revision(self, revision):
        # type: (Revision) -> [Specification]

        assert not revision or isinstance(revision, Revision)

        # logger.debug('Getting cartfile from revision {} of {})'.format(revision, self))

        if revision in self.specifications_cache:
            return self.specifications_cache[revision]
        elif revision is None and self == self.punic.root_project:
            cartfile = Cartfile(use_ssh=self.config.use_ssh, overrides=config.repo_overrides)
            specifications = []

            cartfile_path = self.path / 'Cartfile'
            cartfile_private_path = self.path / 'Cartfile.private'

            if cartfile_path.exists():
                cartfile.read(cartfile_path)
                specifications += cartfile.specifications

            if cartfile_private_path.exists():
                cartfile.read(cartfile_private_path)
                if set(specifications).intersection(cartfile.specifications):
                    raise PunicRepresentableError(
                        "Specifications in your Cartfile.private conflict with specifications within your Cartfile.")
                specifications += cartfile.specifications

            if not specifications:
                raise PunicRepresentableError(
                    "No specifications found in {} or {}".format(cartfile_path.relative_to(Path.cwd()), cartfile_private_path.relative_to(Path.cwd())))

        else:
            self.check_work_directory()

            try:
                parsed_revision = self.rev_parse(revision)
            except:
                print("FAILED") # JIWTODO
                return []

            result = runner.run('git show "{}:Cartfile"'.format(parsed_revision), cwd=self.path)
            if result.return_code != 0:
                specifications = []
            else:
                data = result.stdout
                cartfile = Cartfile(use_ssh=self.config.use_ssh, overrides=config.repo_overrides)
                cartfile.read(data)
                specifications = cartfile.specifications

        self.specifications_cache[revision] = specifications
        return specifications

    def revisions_for_predicate(self, predicate):
        # type: (VersionPredicate) -> [Tag]
        return [tag for tag in self.tags if predicate.test(tag.semantic_version)]


########################################################################################################################

@functools.total_ordering
class Revision(object):
    always_use_is_ancestor = False

    class Type(Enum):
        tag = 'tag'
        commitish = 'commitish'

    def __init__(self, repository, revision, revision_type, check = True):
        assert isinstance(repository, Repository)
        assert isinstance(revision, six.string_types)
        #        assert isinstance(revision_type, Revision.Type) # TODO: This doesn't work.
        self.repository = repository
        self.revision = revision
        self.revision_type = revision_type
        self.semantic_version = (SemanticVersion.string(self.revision) if self.revision_type == Revision.Type.tag else None)

        if check and not self.exists:
            raise NoSuchRevision(repository=repository, revision=revision)

    @mproperty
    def exists(self):
        # TODO: dont use exceptions for logic
        try:
            self.repository.rev_parse(self.revision)
        except:
            return False
        return True


    @mproperty
    def sha(self):
        assert self.repository
        return self.repository.rev_parse(self.revision)

    def __repr__(self):
        return str(self.revision)

    def __eq__(self, other):
        try:
            if self.semantic_version and other.semantic_version and not Revision.always_use_is_ancestor:
                return self.semantic_version == other.semantic_version
            else:
                return (self.revision, self.revision_type) == (other.revision, other.revision_type)
        except:
            logging.error(self.repository)
            raise

    # see: https://bugs.python.org/issue25732
    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        if self.semantic_version and other.semantic_version and not Revision.always_use_is_ancestor:
            return self.semantic_version < other.semantic_version
        else:
            self.repository.check_work_directory()

            result = runner.run('git merge-base --is-ancestor "{}" "{}"'.format(other.sha, self.sha), cwd=self.repository.path)
            if result.return_code == 0:
                return False
            if result.return_code == 1:
                return True
            else:
                raise Exception('git merge-base returned {} {} {} {} {}'.format(result.return_code, result.stderr, self.repository, other.sha, self.sha))

    def __hash__(self):
        # TODO: Should include repo too?
        return hash((self.revision, self.revision_type))



from __future__ import division, absolute_import, print_function

__all__ = ['work_directory', 'timeit']

import contextlib
import os
import time
import logging
import punic

@contextlib.contextmanager
def work_directory(path):
    # type: (Union[Path, None])
    saved_wd = None
    if path:
        path = str(path)
        saved_wd = os.getcwd()
        os.chdir(path)
    try:
        yield
    except:
        raise
    finally:
        if saved_wd:
            os.chdir(saved_wd)


@contextlib.contextmanager
def timeit(task=None, log=None):

    if log is None:
        log = punic.current_session.config.log_timings

    # type: (Union[str, None])
    start = time.time()
    yield
    end = time.time()
    if log:
        logging.info('Task \'<ref>{}</ref>\' took <echo>{:.6f}</echo> seconds.'.format(task if task else '<unnamed task>', end - start))



__all__ = ['GitHub']

import click
import requests
# from prettytable import PrettyTable
from pprint import pprint
from jsonpath_rw import jsonpath, parse
# import csv
# import iso8601
# import sys
# import re
from memoize import mproperty
import subprocess
import shlex

def value(path, json):
    path = parse(path)
    result = path.find(json)
    return result[0].value if result else None

def to_string(o):
    if o is None:
        return ''
    else:
        return str(o)



class GitHub(object):

    def __init__(self):
        self.verbose = False

    @mproperty
    def auth(self):
        try:
            github_user = unicode(subprocess.check_output(shlex.split('git config --get github.user')).strip(), 'utf-8')
            github_token = unicode(subprocess.check_output(shlex.split('git config --get github.token')).strip(), 'utf-8')
        except subprocess.CalledProcessError as e:
            return None
        return (github_user, github_token)

    def search(self, name, language = 'swift'):
        q = '{} language:{} in:name fork:false'.format(name, language)
        headers = {'Accept': 'application/vnd.github.drax-preview+json'}
        r = requests.get('https://api.github.com/search/repositories?q={}&sort=stars'.format(q), auth = self.auth, headers = headers)
        items = r.json()['items']

        repositories = [Repository(github=self, json=item) for item in items]

        return repositories

class Repository(object):

    def __init__(self, github, json):
        self.github = github
        self.json = json
        self.full_name = self.json['full_name']
        self.name = self.json['name']
        self.owner = value('owner.login', self.json)
        self.stargazers_count = self.json['stargazers_count']

    @mproperty
    def license(self):
        headers = {'Accept': 'application/vnd.github.drax-preview+json'}
        r = requests.get('https://api.github.com/repos/{}'.format(self.full_name), auth=self.github.auth, headers=headers)
        return value('license.name', r.json())

    def __repr__(self):
        return self.full_name

# r = requests.get('https://api.github.com/repos/{}/git/refs/tags'.format(first_item['full_name']))
#
# matches = [re.match(r'^refs/tags/(.*)$', row['ref']) for row in r.json()]
# tags = [match.groups()[0] for match in matches if match]
# print(tags)
#
#
#
# sys.exit()
#
#
# r = requests.get('https://api.github.com/repos/schwa/SwiftGraphics/issues', auth=('schwa', ''))
#
# json = r.json()
#
# # writer = csv.writer(open('/Users/schwa/Desktop/issues.csv', 'w'))
# # table = PrettyTable(['Number', 'State', 'Assignee', 'Title'])
#
#
#
# class Issue(object):
#     def __init__(self, json):
#         self.json = json
#
#         self.number = json['number']
#         self.state = json['state']
#         self.title = json['title']
#         self.assignee = value('assignee.login', json)
#         self.body = json['body']
#         self.created = iso8601.parse_date(json['created_at'])
#         self.user = value('user.login', json)
#
#     def markdown(self):
#         d = self.__dict__
#         lines = []
#         lines += ['# {title} (Issue: #{number})'.format(**d), '']
#         lines += ['## Metadata', '']
#         lines += ['State', ':\t{state}'.format(**d), '']
#         lines += ['Created At', ':\t{created}'.format(**d), '']
#         lines += ['Created By', ':\t@{user}'.format(**d), '']
#         if self.assignee:
#             lines += ['Assignee', ':\t@{assignee}'.format(**d)]
#
#         if self.body:
#             lines += ['']
#             lines += ['## Body', '']
#             lines += [self.body]
#
#         return '\n'.join(lines)
#
# issues = [Issue(row) for row in json]
#
# for issue in issues:
#     print(issue.markdown())
#     print()
#     print('*' * 80)
#     print()
#
# #pprint(json[1])



from __future__ import division, absolute_import, print_function

__all__ = ['copy_frameworks_main']

import os
import re
from pathlib2 import Path
import logging

from .runner import *
from .xcode import uuids_from_binary
import punic.shshutil as shutil


def copy_frameworks_main():
    sym_root = Path(os.environ['SYMROOT'])
    valid_architectures = set(os.environ['VALID_ARCHS'].split(' '))
    input_file_count = int(os.environ['SCRIPT_INPUT_FILE_COUNT'])
    input_files = [Path(os.environ.get('SCRIPT_INPUT_FILE_{}'.format(index))) for index in range(0, input_file_count)]
    expanded_identity = os.environ['EXPANDED_CODE_SIGN_IDENTITY_NAME']
    built_products_dir = Path(os.environ['BUILT_PRODUCTS_DIR'])
    frameworks_folder_path = os.environ['FRAMEWORKS_FOLDER_PATH']
    frameworks_path = built_products_dir / frameworks_folder_path
    code_signing_allowed = os.environ['CODE_SIGNING_ALLOWED'] == 'YES'
    enable_bitcode = os.environ['ENABLE_BITCODE'] == 'YES'
    project_dir = Path(os.environ['PROJECT_DIR'])
    platform_display_name = os.environ['PLATFORM_DISPLAY_NAME']
    punic_builds_dir = project_dir / 'Carthage' / 'Build' / platform_display_name
    action = os.environ['ACTION']

    for input_path in input_files:

        logging.info('Processing: "{}"'.format(input_path.name))

        # We don't modify the input frameworks but rather the ones in the built products directory
        output_path = frameworks_path / input_path.name

        framework_name = input_path.stem

        logging.info('\tCopying framework "{}" to "$SYMROOT/{}"'.format(framework_name, output_path.relative_to(sym_root)))
        if output_path.exists():
            shutil.rmtree(output_path)

        def ignore(src, names):
            src = Path(src)
            if src.suffix == ".framework":
                return ['Headers', 'PrivateHeaders', 'Modules']
            else:
                return []

        shutil.copytree(input_path, output_path, symlinks=True, ignore = ignore)

        framework_path = output_path

        binary_path = framework_path / framework_name

        if code_signing_allowed:
            # Find out what architectures the framework has
            output = runner.check_call(['/usr/bin/xcrun', 'lipo', '-info', binary_path])
            match = re.match(r'^Architectures in the fat file: (.+) are: (.+)'.format(binary_path), output)
            assert match.groups()[0] == str(binary_path)
            architectures = set(match.groups()[1].strip().split(' '))
            logging.info('\tArchitectures: {}'.format(list(architectures)))

            # Produce a list of architectures that are not valid
            excluded_architectures = architectures.difference(valid_architectures)

            # Skip if all architectures are valid
            if not excluded_architectures:
                continue

            # For each invalid architecture strip it from framework
            for architecture in excluded_architectures:
                logging.info('\tStripping "{}" from "{}"'.format(architecture, framework_name))
                output = runner.check_call(['/usr/bin/xcrun', 'lipo', '-remove', architecture, '-output', binary_path, binary_path])

                # Resign framework
                logging.info('\tResigning "{}"/"{}" with "{}"'.format(framework_name, architecture, expanded_identity))

            logging.info('\tCode signing: "$SYMROOT/{}"'.format(binary_path.relative_to(sym_root)))

            # noinspection PyUnusedLocal
            result = runner.check_call(['/usr/bin/xcrun', 'codesign', '--force', '--sign', expanded_identity, '--preserve-metadata=identifier,entitlements', binary_path])
        else:
            logging.info('\tCode signing not allowed. Skipping.')

        if action == 'install':
            uuids = uuids_from_binary(binary_path)

            # Copy dSYM files from $PROJECT_DIRCarthage/Build to $BUILT_PRODUCTS_DIR
            dsym_path = input_path.parent / (binary_path.name + '.dSYM')

            if dsym_path.exists():

                dsym_output_path = built_products_dir / dsym_path.name

                logging.info('\tCopying "$PROJECT_DIR/{}" to "$BUILT_PRODUCTS_DIR"'.format(dsym_path.relative_to(project_dir)))
                if dsym_output_path.exists():
                    shutil.rmtree(dsym_output_path)
                shutil.copytree(dsym_path, dsym_output_path, symlinks=True)

            # Copy bcsymbolmap files from $PROJECT_DIRCarthage/Build to $BUILT_PRODUCTS_DIR
            if enable_bitcode:
                for uuid in uuids:
                    bcsymbolmap_path = punic_builds_dir / (uuid + '.bcsymbolmap')
                    logging.info('\tCopying "$PROJECT_DIR/{}" to "$BUILT_PRODUCTS_DIR"'.format(bcsymbolmap_path.relative_to(project_dir)))
                    shutil.copy(bcsymbolmap_path, built_products_dir)



from __future__ import division, absolute_import, print_function

__all__ = ['Specification', 'ProjectIdentifier', 'VersionOperator', 'VersionPredicate']

import functools
import re
import six
from flufl.enum import Enum
from memoize import mproperty
from pathlib2 import Path
import logging

from .semantic_version import *

# Ideally we could six.urllib but this causes problem with nosetests!
if six.PY2:
    import urlparse
elif six.PY3:
    import urllib.parse as urlparse


class Specification(object):
    def __init__(self, identifier, predicate = None):
        # type: (str, VersionPredicate)
        self.identifier = identifier
        self.predicate = predicate
        self.raw_string = None

    @classmethod
    def cartfile_string(cls, string, use_ssh=False, overrides=None):
        # type: (str) -> Specification
        """
        >>> Specification.cartfile_string('github "foo/bar"')
        github "foo/bar"
        >>> Specification.cartfile_string('github "foo/bar" "master"').identifier
        foo/bar
        >>> Specification.cartfile_string('github "foo/bar" "master"').predicate
        "master"
        >>> Specification.cartfile_string('github "foo/bar" "master"')
        github "foo/bar" "master"
        >>> Specification.cartfile_string('github "foo/bar" >= 1.0').predicate
        >= 1.0
        >>> Specification.cartfile_string('github "ReactiveX/RxSwift" "some/branch"').identifier
        ReactiveX/RxSwift
        >>> Specification.cartfile_string('github "ReactiveX/RxSwift" "some/branch"').predicate
        "some/branch"
        >>> Specification.cartfile_string('git "file:///Users/example/Project" "some/branch"').identifier
        Project
        >>> Specification.cartfile_string('git "git@gitlab.com:mokagio/punic-cartfile-issue.git" "master"').identifier
        punic-cartfile-issue
        """

        match = re.match(r'^(?P<address>(?P<service>github|git)\s+"[^/]+/(?:.+?)")(?:\s+(?P<predicate>.+)?)?', string)
        if not match:
            raise Exception('Bad spec {}'.format(string))

        identifier = ProjectIdentifier.string(match.group('address'), use_ssh=use_ssh, overrides=overrides)
        predicate = VersionPredicate(match.group('predicate'))
        specification = Specification(identifier=identifier, predicate=predicate)
        specification.raw_string = string

        return specification

    def __repr__(self):
        parts = [self.identifier.full_identifier]
        if self.predicate:
            parts += [str(self.predicate)]
        parts = [part for part in parts if part]
        return ' '.join(parts)


@functools.total_ordering
class ProjectIdentifier(object):
    @classmethod
    def string(cls, string, use_ssh=False, overrides=None):
        # type: (str) -> ProjectIdentifier
        """
        >>> ProjectIdentifier.string('github "foo/bar"')
        foo/bar
        >>> ProjectIdentifier.string('github "foo/bar"').team_name
        'foo'
        >>> ProjectIdentifier.string('github "foo/bar"').project_name
        'bar'
        >>> ProjectIdentifier.string('github "foo/bar"').identifier
        'foo/bar'
        >>> ProjectIdentifier.string('github "foo/bar"').full_identifier
        'github "foo/bar"'
        >>> ProjectIdentifier.string('git "file:///Users/example/Projects/Example-Project"')
        Example-Project
        >>> ProjectIdentifier.string('git "git@gitlab.com:mokagio/punic-cartfile-issue.git"')
        punic-cartfile-issue
        """

        assert isinstance(string, six.string_types)
        assert isinstance(use_ssh, bool)

        match = re.match(r'^(?P<source>github|git)\s+"(?P<link>.+)"', string)
        if not match:
            raise Exception('No match')

        source = match.group('source')
        link = match.group('link')

        if source == 'github':
            match = re.match(r'^(?P<team_name>[^/]+)/(?P<project_name>[^/]+)$', link)
            if not match:
                raise Exception('No match')
            team_name = match.group('team_name')
            project_name = match.group('project_name')

            if not use_ssh:
                remote_url = 'https://github.com/{}/{}.git'.format(team_name, project_name)
            else:
                remote_url = 'git@github.com:{}/{}.git'.format(team_name, project_name)
        elif source == 'git':
            team_name = None
            url_parts = urlparse.urlparse(link)
            path = Path(url_parts.path)
            project_name = path.stem
            remote_url = link
        else:
            raise Exception('No match')

        return ProjectIdentifier(source=source, remote_url=remote_url, team_name=team_name, project_name=project_name,
            overrides=overrides)

    def __init__(self, source=None, team_name=None, project_name=None, remote_url=None, overrides=None):
        self.source = source
        self.team_name = team_name
        self.project_name = project_name
        self.remote_url = remote_url
        if overrides and self.project_name in overrides:
            override_url = overrides[self.project_name]
            logging.info('Overriding {} with git URL {}'.format(self.project_name, override_url))
            self.remote_url = override_url

    @mproperty
    def full_identifier(self):
        if self.source == 'git':
            return '{} "{}"'.format(self.source, self.remote_url)
        elif self.source == 'github':
            return '{} "{}/{}"'.format(self.source, self.team_name, self.project_name)
        else:
            raise Exception("Unknown source")

    @mproperty
    def identifier(self):
        components = [] \
                     + ([self.team_name] if self.team_name else []) \
                     + [self.project_name]
        return '/'.join(components)

    def __repr__(self):
        return self.identifier

    def __eq__(self, other):
        """
        >>> ProjectIdentifier.string('github "foo/bar"') == ProjectIdentifier.string('github "foo/bar"')
        True
        """
        return self.identifier.lower() == other.identifier.lower()

    def __ne__(self, other):
        """
        >>> ProjectIdentifier.string('github "foo/bar"') != ProjectIdentifier.string('github "foo/bar2"')
        True
        """
        return not (self == other)

    def __lt__(self, other):
        """
        >>> ProjectIdentifier.string('github "foo/bar"') < ProjectIdentifier.string('github "foo/bar2"')
        True
        """
        return self.identifier.lower() < other.identifier.lower()

    def __hash__(self):
        """
        >>> hash(ProjectIdentifier.string('github "foo/bar"')) == hash(ProjectIdentifier.string('github "foo/bar"'))
        True
        >>> hash(ProjectIdentifier.string('github "foo/bar"')) != hash(ProjectIdentifier.string('github "foo/bar2"'))
        True
        """
        return hash(self.identifier.lower())

    def matches(self, name_filter):
        # type: ([str]) -> bool

        if not name_filter:
            return True

        if self.project_name in name_filter:
            return True

        return False


class VersionOperator(Enum):
    commitish = 'commit-ish'
    any = '<any>'
    greater_than_or_equals = '>='
    equals = '=='
    semantic_greater_than_or_equals = '~>'


class VersionPredicate(object):
    def __init__(self, string):
        """
        >>> VersionPredicate('"master"')
        "master"
        >>> VersionPredicate('>= 1.0')
        >= 1.0
        >>> VersionPredicate('~> 1.0')
        ~> 1.0
        >>> VersionPredicate('== 1.0')
        == 1.0
        """

        if not string:
            self.operator = VersionOperator.any
            self.value = None
        else:
            match = re.match(r'(?:(~>|>=|==|)\s+)?(?:"(.+)"|(.+))', string)
            if not match:
                raise Exception('No match for: {}'.format(string))

            operator = match.group(1)
            value = match.group(2) if match.group(2) else match.group(3)

            if operator == '==':
                self.operator = VersionOperator.equals
                self.value = SemanticVersion.string(value)
            elif operator == '>=':
                self.operator = VersionOperator.greater_than_or_equals
                self.value = SemanticVersion.string(value)
            elif operator == '~>':
                self.operator = VersionOperator.semantic_greater_than_or_equals
                self.value = SemanticVersion.string(value)
            else:
                self.operator = VersionOperator.commitish
                self.value = value

    def __repr__(self):
        if self.operator == VersionOperator.any:
            return ''
        if self.operator == VersionOperator.commitish:
            return '"{}"'.format(self.value)
        elif self.operator == VersionOperator.equals:
            return '== {}'.format(self.value)
        elif self.operator == VersionOperator.greater_than_or_equals:
            return '>= {}'.format(self.value)
        elif self.operator == VersionOperator.semantic_greater_than_or_equals:
            return '~> {}'.format(self.value)

    def test(self, version):
        """
        >>> VersionPredicate('== 1.0').test(SemanticVersion.string('1.0'))
        True
        >>> VersionPredicate('== 1.0').test(SemanticVersion.string('1.1'))
        False
        >>> VersionPredicate('>= 1.0').test(SemanticVersion.string('1.1'))
        True
        >>> VersionPredicate('>= 1.1').test(SemanticVersion.string('1.0'))
        False
        >>> VersionPredicate('~> 1.0').test(SemanticVersion.string('1.0'))
        True
        >>> VersionPredicate('~> 1.0').test(SemanticVersion.string('0.9'))
        False
        >>> VersionPredicate('~> 1.0').test(SemanticVersion.string('1.0.1'))
        True
        >>> VersionPredicate('~> 1.0').test(SemanticVersion.string('1.1'))
        False
        >>> VersionPredicate('~> 1.0').test(SemanticVersion.string('2.0'))
        False
        """

        # type: (SemanticVersion) -> bool
        if self.operator == VersionOperator.any:
            return True
        elif self.operator == VersionOperator.equals:
            return version == self.value
        elif self.operator == VersionOperator.greater_than_or_equals:
            return version >= self.value
        elif self.operator == VersionOperator.semantic_greater_than_or_equals:
            return self.value <= version < self.value.next_minor
        return False



import os
import re
import logging
import itertools

from punic import shshutil as shutil
from punic.config import config
from punic.runner import runner
from punic.xcode import XcodeProject


class Checkout(object):

    def __init__(self, punic, identifier, revision):
        self.punic = punic
        self.identifier = identifier
        self.repository = self.punic._repository_for_identifier(self.identifier)
        self.revision = revision
        self.checkout_path = self.config.checkouts_path / self.identifier.project_name

    @property
    def config(self):
        return self.punic.config

    def prepare(self):

        if self.config.use_submodules:
            relative_checkout_path = self.checkout_path.relative_to(self.config.root_path)

            result = runner.run('git submodule status "{}"'.format(relative_checkout_path))
            if result.return_code == 0:
                match = re.match(r'^(?P<flag> |\-|\+|U)(?P<sha>[a-f0-9]+) (?P<path>.+)( \((?P<description>.+)\))?', result.stdout)
                flag = match.groupdict()['flag']
                if flag == ' ':
                    pass
                elif flag == '-':
                    raise Exception('Uninitialized submodule {}. Please report this!'.format(self.checkout_path))
                elif flag == '+':
                    raise Exception('Submodule {} doesn\'t match expected revision'.format(self.checkout_path))
                elif flag == 'U':
                    raise Exception('Submodule {} has merge conflicts'.format(self.checkout_path))
            else:
                if self.checkout_path.exists():
                    raise Exception('Want to create a submodule in {} but something already exists in there.'.format(self.checkout_path))
                logging.debug('Adding submodule for {}'.format(self))
                runner.check_run(['git', 'submodule', 'add', '--force', self.identifier.remote_url, self.checkout_path.relative_to(self.config.root_path)])

            # runner.check_run(['git', 'submodule', 'add', '--force', self.identifier.remote_url, self.checkout_path.relative_to(self.config.root_path)])
            # runner.check_run(['git', 'submodule', 'update', self.checkout_path.relative_to(self.config.root_path)])

            logging.debug('Updating {}'.format(self))
            self.repository.checkout(self.revision)
        else:

            # TODO: This isn't really 'fetch'
            if self.config.fetch:

                self.repository.checkout(self.revision)
                logging.debug('<sub>Copying project to <ref>Carthage/Checkouts</ref></sub>')
                if self.checkout_path.exists():
                    shutil.rmtree(self.checkout_path, ignore_errors=True)
                shutil.copytree(self.repository.path, self.checkout_path, symlinks=True, ignore=shutil.ignore_patterns('.git'))

        if not self.checkout_path.exists():
            raise Exception('No checkout at path: {}'.format(self.checkout_path))

        # We only need to bother making a symlink to <root>/Carthage/Build if dependency also has dependencies.
        if len(self.punic.dependencies_for_project_and_tag(self.identifier, self.revision)):
            # Make a Carthage/Build symlink inside checked out project.
            carthage_path = self.checkout_path / 'Carthage'
            if not carthage_path.exists():
                carthage_path.mkdir()

            carthage_symlink_path = carthage_path / 'Build'
            if carthage_symlink_path.exists():
                carthage_symlink_path.unlink()
            logging.debug('<sub>Creating symlink: <ref>{}</ref> to <ref>{}</ref></sub>'.format(carthage_symlink_path.relative_to(self.config.root_path), self.config.build_path.relative_to(self.config.root_path)))
            assert self.config.build_path.exists()

            # TODO: Generate this programatically.
            os.symlink("../../../Build", str(carthage_symlink_path))

    @property
    def projects(self):
        def _make_cache_identifier(project_path):
            rev = self.repository.rev_parse(self.revision)
            cache_identifier = '{},{}'.format(str(rev), project_path.relative_to(self.checkout_path))
            return cache_identifier

        def test(path):
            relative_path = path.relative_to(self.checkout_path)
            if 'Carthage/Checkouts' in str(relative_path):
                return False
            return True

        project_paths = itertools.chain(self.checkout_path.glob("**/*.xcworkspace"), self.checkout_path.glob("**/*.xcodeproj"))
        project_paths = [path for path in project_paths if test(path)]
        if not project_paths:
            logging.warning("No projects/workspaces found in {}".format(self.checkout_path))
            return []

        projects = []
        schemes = []
        embedded_project_pattern = re.compile(r"\.(playground|xcodeproj)/[^/]+\.xcworkspace$")

        for project_path in project_paths:
            if embedded_project_pattern .search(str(project_path)):
                #print('Skipping project\'s embedded workspace:', project_path)
                continue
            project = XcodeProject(self, config.xcode, project_path, _make_cache_identifier(project_path))
            for scheme in list(project.scheme_names):
                if scheme in schemes:
                    project.info[2].remove(scheme)
                else:
                    schemes.append(scheme)
            if len(project.scheme_names) > 0:
                projects.append(project)
        return projects



__all__ = ['github_search']

import logging
from pathlib2 import Path

from punic.styling import styled
from punic.github import *
from punic.cartfile import *
from punic.specification import *

import six

def menu(prompt, items, formatter=None, default=None):
    formatter = formatter or str
    for index, item in enumerate(items):
        print('{}: {}'.format(index + 1, formatter(item)))
    while True:
        s = raw_input('>')
        if not s and default is not None:
            return default
        else:
            try:
                index = int(s)
                return items[index - 1]
            except:
                continue

def github_search(punic, name, cartfile_append = True, language='swift'):
    repositories = GitHub().search(name, language = language)

    logging.info('Found {} repositories matching \'{}\'. Filtering...'.format(len(repositories), name))


    # Get rid of forks.
    repositories = [repo for repo in repositories if not repo.json['fork']]

    # Get rid of zero stars.
    repositories = [repo for repo in repositories if repo.json['stargazers_count']]

    # Limit to 10 items
    repositories = repositories[:10]

    def formatter(repo):
        s = '<ref>{repo.full_name}</ref>, stars: {repo.stargazers_count}, license: {repo.license}'.format(repo=repo)
        return styled(s)

    if not cartfile_append:

        for repository in repositories:
            logging.info(formatter(repository))

    else:
        # Get rid of no license.
        repositories = [repo for repo in repositories if repo.license]

        repository = menu('?', repositories, formatter)

        repository = repositories[0]
        append_to_cartfile(punic, repository)

def append_to_cartfile(punic, repository):

    cartfile_path = punic.config.root_path / 'Cartfile'
    if cartfile_path.exists():
        cartfile = Cartfile()
        specifications = cartfile.read(cartfile_path)
    else:
        specifications = []

    project_identifier = ProjectIdentifier(source='github', team_name=repository.owner, project_name=repository.name)
    for specification in specifications:
        if specification.identifier == project_identifier:
            logging.warning('Project \'<ref>{}</ref>\' already in cartfile'.format(project_identifier))
            return

    logging.info('Adding \'<ref>{}</ref>\' to Cartfile'.format(project_identifier))

    new_specification = Specification(identifier=project_identifier, predicate=None)
    cartfile = Cartfile(specifications + [new_specification])
    cartfile.write(cartfile_path.open('w'))



from __future__ import division, absolute_import, print_function

__all__ = ['punic_cli', 'main']

import logging
import logging.handlers
import sys
import click
from click_didyoumean import DYMGroup
import networkx as nx
from pathlib2 import Path

import punic
from .copy_frameworks import *
from .errors import *
from .logger import *
from .semantic_version import *
from .utilities import *
from .version_check import *
from .config_init import *
from .carthage_cache import *
from punic.graph import make_graph
import punic.shshutil as shutil
from punic import *
from .runner import *
from .checkout import *
from .search import *

@click.group(cls=DYMGroup)
@click.option('--echo', default=False, is_flag=True, help="""Echo all commands to terminal.""")
@click.option('--verbose', default=False, is_flag=True, help="""Verbose logging.""")
@click.option('--color/--no-color', default=None, is_flag=True, help="""TECHNICOLOR.""")
@click.option('--timing/--no-timing', default=False, is_flag=True, help="""Log timing info""")
@click.pass_context
def punic_cli(context, echo, verbose, timing, color):
    ### TODO: Clean this up!

    # Configure click
    context.token_normalize_func = lambda x: x if not x else x.lower()

    # Configure logging
    level = logging.DEBUG if verbose else logging.INFO

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    formatter = HTMLFormatter()

    # create console handler and set level to debug
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(stream_handler)

    # TODO: This needs to be a better location
    logs_path = Path('~/Library/Application Support/io.schwa.Punic/Logs').expanduser()
    if not logs_path.exists():
        logs_path.mkdir(parents=True)

    log_path = logs_path / "punic.log"
    needs_rollover = log_path.exists()

    file_handler = logging.handlers.RotatingFileHandler(str(log_path), backupCount=4)
    if needs_rollover:
        file_handler.doRollover()
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(HTMLStripperFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")))
    logger.addHandler(file_handler)

    for name in ['boto', 'requests.packages.urllib3']:
        named_logger = logging.getLogger(name)
        named_logger.setLevel(logging.WARNING)
        named_logger.propagate = True

    runner.echo = echo

    # Set up punic
    punic = Punic()
    punic.config.log_timings = timing
    context.obj = punic
    punic.config.verbose = verbose
    punic.config.echo = verbose

    # Color:
    if color is None:
        if punic.config.continuous_integration:
            color = False

    if color is None:
        color = True


    punic.config.color = color
    formatter.color = color
    logger.color = color



@punic_cli.command()
@click.pass_context
@click.option('--use-submodules', default=None, help="""Add dependencies as Git submodules""")
@click.option('--use-ssh', default=None, help="""Use SSH for downloading GitHub repositories""")
def fetch(context, **kwargs):
    """Fetch the project's dependencies.."""
    logging.info("<cmd>fetch</cmd>")
    punic = context.obj
    punic.config.fetch = True  # obviously
    punic.config.update(**kwargs)

    with timeit('fetch'):
        with error_handling():
            punic.fetch()


@punic_cli.command()
@click.pass_context
@click.option('--fetch/--no-fetch', default=True, is_flag=True, help="""Controls whether to fetch dependencies.""")
@click.option('--use-submodules', default=None, help="""Add dependencies as Git submodules""")
@click.option('--use-ssh', default=None, is_flag=True, help="""Use SSH for downloading GitHub repositories""")
def resolve(context, **kwargs):
    """Resolve dependencies and output `Carthage.resolved` file.

    This sub-command does not build dependencies. Use this sub-command when a dependency has changed and you just want to update `Cartfile.resolved`.
    """
    punic = context.obj
    logging.info("<cmd>Resolve</cmd>")
    punic.config.update(**kwargs)


    with timeit('resolve'):
        with error_handling():
            punic.resolve()


@punic_cli.command()
@click.pass_context
@click.option('--configuration', default=None, help="""Dependency configurations to build. Usually 'Release' or 'Debug'.""")
@click.option('--platform', default=None, help="""Platform to build. Comma separated list.""")
@click.option('--fetch/--no-fetch', default=True, is_flag=True, help="""Controls whether to fetch dependencies.""")
@click.option('--xcode-version', default=None, help="""Xcode version to use""")
@click.option('--toolchain', default=None, help="""Xcode toolchain to use""")
@click.option('--dry-run', default=None, is_flag=True, help="""Do not actually perform final build""")
@click.option('--use-submodules', default=None, help="""Add dependencies as Git submodules""")
@click.option('--use-ssh', default=None, is_flag=True, help="""Use SSH for downloading GitHub repositories""")
@click.argument('deps', nargs=-1)
def build(context, **kwargs):
    """Fetch and build the project's dependencies."""
    logging.info("<cmd>Build</cmd>")
    punic = context.obj
    punic.config.update(**kwargs)

    deps = kwargs['deps']

    logging.debug('Platforms: {}'.format(punic.config.platforms))
    logging.debug('Configuration: {}'.format(punic.config.configuration))

    with timeit('build'):
        with error_handling():
            punic.build(dependencies=deps)


@punic_cli.command()
@click.pass_context
@click.option('--configuration', default=None, help="""Dependency configurations to build. Usually 'Release' or 'Debug'.""")
@click.option('--platform', default=None, help="""Platform to build. Comma separated list.""")
@click.option('--fetch/--no-fetch', default=True, is_flag=True, help="""Controls whether to fetch dependencies.""")
@click.option('--xcode-version', default=None, help="""Xcode version to use""")
@click.option('--toolchain', default=None, help="""Xcode toolchain to use""")
@click.option('--use-submodules', default=None, help="""Add dependencies as Git submodules""")
@click.option('--use-ssh', default=None, is_flag=True, help="""Use SSH for downloading GitHub repositories""")
@click.argument('deps', nargs=-1)
def update(context, **kwargs):
    """Update and rebuild the project's dependencies."""
    logging.info("<cmd>Update</cmd>")
    punic = context.obj
    punic.config.update(**kwargs)

    deps = kwargs['deps']

    with timeit('update'):
        with error_handling():
            punic.resolve()
            punic.build(dependencies=deps)


@punic_cli.command()
@click.pass_context
@click.option('--derived-data', default=False, is_flag=True, help="""Clean the punic derived data directory.""")
@click.option('--caches', default=False, is_flag=True, help="""Clean the global punic files.""")
@click.option('--build', default=False, is_flag=True, help="""Clean the locate Carthage/Build directorys.""")
@click.option('--all', default=False, is_flag=True, help="""Clean all.""")
def clean(context, derived_data, caches, build, all):
    """Clean project & punic environment."""
    logging.info("<cmd>Clean</cmd>")
    punic = context.obj

    with timeit('clean'):
        if build or all:
            logging.info('Erasing Carthage/Build directory')
            if punic.config.build_path.exists():
                shutil.rmtree(punic.config.build_path)

        if derived_data or all:
            logging.info('Erasing derived data directory')
            if punic.config.derived_data_path.exists():
                shutil.rmtree(punic.config.derived_data_path)

        if caches or all:
            if punic.config.repo_cache_directory.exists():
                logging.info('Erasing {}'.format(punic.config.repo_cache_directory))
                shutil.rmtree(punic.config.repo_cache_directory)
            logging.info('Erasing run cache')
            runner.reset()


@punic_cli.command()
@click.pass_context
@click.option('--fetch/--no-fetch', default=True, is_flag=True, help="""Controls whether to fetch dependencies.""")
@click.option('--use-submodules', default=None, help="""Add dependencies as Git submodules""")
@click.option('--use-ssh', default=None, is_flag=True, help="""Use SSH for downloading GitHub repositories""")
@click.option('--open', default=False, is_flag=True, help="""Open the graph image file.""")
def graph(context, fetch, use_submodules, use_ssh, open):
    """Output resolved dependency graph."""
    logging.info("<cmd>Graph</cmd>")
    punic = context.obj
    punic.config.fetch = fetch
    if use_submodules:
        punic.config.use_submodules = use_submodules
    if use_ssh:
        punic.config.use_ssh = use_ssh
    with timeit('graph'):
        make_graph(punic, open)


@punic_cli.command(name='copy-frameworks')
@click.pass_context
def copy_frameworks(context):
    """In a Run Script build phase, copies each framework specified by a SCRIPT_INPUT_FILE environment variable into the built app bundle."""
    copy_frameworks_main()


# noinspection PyUnusedLocal
@punic_cli.command()
@click.pass_context
@click.option('--check/--no-check', default=True, help="""Check for latest version.""")
@click.option('--simple', is_flag=True, default=False, help="""Only display simple version info. Implies --no-check.""")
@click.option('--xcode', is_flag=True, default=False, help="""Display xcode versions.""")
def version(context, check, simple, xcode):
    """Display the current version of Punic."""

    from .styling import styled_print

    if simple:
        print("{}".format(punic.__version__))
    else:

        styled_print('Punic version: <version>{}</version>'.format(punic.__version__))

        if punic.config.verbose:
            styled_print('Punic path: <ref>{}</ref> '.format(punic.__file__))

        sys_version = sys.version_info
        sys_version = SemanticVersion.from_dict(dict(major=sys_version.major, minor=sys_version.minor, micro=sys_version.micro, releaselevel=sys_version.releaselevel, serial=sys_version.serial, ))
        styled_print('Python version: <version>{}</version>'.format(sys_version))
        if punic.config.verbose:
            styled_print('Python path: <path>{}</path> '.format(sys.executable))

        if xcode or punic.config.verbose:
            styled_print("Xcode(s):")
            for xcode in punic.xcode.Xcode.find_all():
                s = '\t<path>{}</path>: <version>{}</version>'.format(xcode.path, xcode.version)
                if xcode.is_default:
                    s += ' (default)'
                    styled_print(s)

        if check:
            version_check(verbose=True, timeout=None, failure_is_an_option=False)


@punic_cli.command()
@click.pass_context
def readme(context):
    """Opens punic readme in your browser (https://github.com/schwa/punic/blob/HEAD/README.markdown)"""
    click.launch('https://github.com/schwa/punic/blob/HEAD/README.markdown')


@punic_cli.command()
@click.option('--configuration', default=None, help="""Dependency configurations to build. Usually 'Release' or 'Debug'.""")
@click.option('--platform', default=None, help="""Platform to build. Comma separated list.""")
@click.option('--fetch/--no-fetch', default=True, is_flag=True, help="""Controls whether to fetch dependencies.""")
@click.option('--xcode-version', default=None, help="""Xcode version to use""")
@click.option('--toolchain', default=None, help="""Xcode toolchain to use""")
@click.option('--use-submodules', default=None, help="""Add dependencies as Git submodules""")
@click.option('--use-ssh', default=None, is_flag=True, help="""Use SSH for downloading GitHub repositories""")
@click.argument('deps', nargs=-1)
@click.pass_context
def list(context, **kwargs):
    """Lists all platforms, projects, xcode projects, schemes for all dependencies."""
    punic = context.obj
    punic.config.update(**kwargs)
    deps = kwargs['deps']

    config = punic.config

    configuration, platforms = config.configuration, config.platforms

    if not config.build_path.exists():
        config.build_path.mkdir(parents=True)

    filtered_dependencies = punic._ordered_dependencies(name_filter=deps)

    checkouts = [Checkout(punic=punic, identifier=identifier, revision=revision) for identifier, revision in filtered_dependencies]

    tree = {}

    for platform in platforms:
        tree[platform.name] = {}
        for checkout in checkouts:
            tree[platform.name][str(checkout.identifier)] = {'projects':{}}
            checkout.prepare()
            for project in checkout.projects:
                tree[platform.name][str(checkout.identifier)]['projects'][project.path.name] = {'schemes':{}}
                tree[platform.name][str(checkout.identifier)]['projects'][project.path.name]['path'] = str(project.path.relative_to(config.checkouts_path))
                schemes = project.schemes
                schemes = [scheme for scheme in schemes if scheme.framework_targets]
                schemes = [scheme for scheme in schemes if platform.device_sdk in scheme.supported_platform_names]
                tree[platform.name][str(checkout.identifier)]['projects'][project.path.name]['schemes'] = [scheme.name for scheme in schemes]

    import yaml

    yaml.safe_dump(tree, stream = sys.stdout)

@punic_cli.command()
@click.pass_context
@click.option('--configuration', default=None, help="""Dependency configurations to build. Usually 'Release' or 'Debug'.""")
@click.option('--platform', default=None, help="""Platform to build. Comma separated list.""")
@click.option('--xcode', default=None)
def init(context, **kwargs):
    """Generate punic configuration file."""
    config_init(**kwargs)


@punic_cli.group(cls=DYMGroup)
@click.pass_context
def cache(context):
    """Cache punic build artifacts to Amazon S3"""
    pass


@cache.command()
@click.pass_context
@click.option('--xcode-version', default=None, help="""Xcode version to use""")
@click.option('--force', default=False, is_flag=True, help="""Force publishing""")
def publish(context, xcode_version, force):
    """Generates and uploads the cache archive for the current Cartfile.resolved"""
    with error_handling():
        logging.info("<cmd>Cache Publish</cmd>")
        punic = context.obj
        if xcode_version:
            punic.config.xcode_version = xcode_version
        carthage_cache = CarthageCache(config=punic.config)
        logging.info("Cache filename: <ref>'{}'</ref>".format(carthage_cache.archive_name_for_project()))
        carthage_cache.publish(force = force)



@cache.command()
@click.pass_context
@click.option('--xcode-version', default=None, help="""Xcode version to use""")
def install(context, xcode_version):
    """Installs the cache archive for the current Cartfile.resolved"""
    with error_handling():
        logging.info("<cmd>Cache Install</cmd>")
        punic = context.obj
        if xcode_version:
            punic.config.xcode_version = xcode_version
        carthage_cache = CarthageCache(config=punic.config)
        logging.info("Cache filename: <ref>'{}'</ref>".format(carthage_cache.archive_name_for_project()))
        carthage_cache.install()



@punic_cli.command()
@click.pass_context
@click.argument('name')
@click.option('--append', is_flag=True, default=False, help="""Add a selected project to a Cartfile""")
@click.option('--language', default='swift', help="""Search for projects of specified language""")
def search(context, name, append, language):
    """Search github for repositories and optionally add them to a Cartfile."""
    punic = context.obj

    github_search(punic, name, cartfile_append=append, language=language)


def main():
    punic_cli()



import shutil
from pathlib2 import Path


def rmtree(path, ignore_errors=False, onerror=None):
    shutil.rmtree(str(path), ignore_errors, onerror)


def copytree(src, dst, symlinks=False, ignore=None):
    shutil.copytree(str(src), str(dst), symlinks, ignore)


def copy(src, dst):
    shutil.copy(str(src), str(dst))


def copyfile(src, dst):
    shutil.copyfile(str(src), str(dst))


def ignore_patterns(*patterns):
    return shutil.ignore_patterns(*patterns)


def move(src, dst):
    shutil.move(str(src), str(dst))



__all__ = ['CarthageCache']

import re
import hashlib
import logging
import zipfile
import boto
import boto.s3
import os
import tempfile
from tqdm import tqdm
import yaml

from .shshutil import *
from .errors import *


class CarthageCache(object):
    def __init__(self, config):
        self.config = config

        config_path = Path('.carthage_cache.yml')
        if not config_path.exists():
            raise Exception('No cache configile at <ref>{}</ref>'.format(config_path))

        d = yaml.safe_load(config_path.open())

        client_options = d[':aws_s3_client_options']
        self.AWS_ACCESS_KEY_ID = client_options[':access_key_id']
        self.AWS_SECRET_ACCESS_KEY = client_options[':secret_access_key']
        self.bucket_name = d[':bucket_name']

    #        self.region = client_options.get(':region', None)

    @property
    def archives_directory_path(self):
        path = self.config.library_directory / "Archives"
        if not path.exists():
            path.mkdir(parents=True)
        return path

    def hash_for_project(self):
        output = self.config.xcode.check_call('swift -version')
        swift_version = re.search(r'Swift version ((?:\d+\.)*(?:\d+))', output).group(1)

        resolve_file = Path('Cartfile.resolved').open().read()
        data = '{}{}'.format(resolve_file, swift_version)
        hash = hashlib.sha256(data).hexdigest()
        return hash

    def archive_name_for_project(self):
        return '{}.zip'.format(self.hash_for_project())

    def archive(self, force = False):

        hash = self.hash_for_project()
        archive_file_name = self.archive_name_for_project()

        archive_path = self.archives_directory_path / archive_file_name
        if archive_path.exists() and not force:
            logging.info('Archive already exists in {}. Not recreating.'.format(self.archives_directory_path))
            return archive_path

        logging.info("Creating zipfile.")

        temp_dir = tempfile.mkdtemp()
        temp_archive_path = Path(temp_dir) / archive_file_name

        with zipfile.ZipFile(str(temp_archive_path), 'w', zipfile.ZIP_DEFLATED) as archive:
            def zipdir(root, ziph):
                all_files = list(walk_directory(root))

                logging.info('Computing total size')
                total_size = sum([file.stat().st_size for file in all_files])

                logging.info('Zipping')
                bar = tqdm(total=total_size, unit='B', unit_scale=True)
                for file in all_files:
                    ziph.write(str(file), str(file.relative_to(root)))
                    bar.update(file.stat().st_size)

            zipdir('Carthage/Build', archive)

        copyfile(temp_archive_path, archive_path)

        return archive_path

    def publish(self, archive_path=None, force=False):
        if not archive_path or force:
            archive_path = self.archive(force = force)

        conn = boto.connect_s3(self.AWS_ACCESS_KEY_ID, self.AWS_SECRET_ACCESS_KEY)
        bucket = conn.get_bucket(self.bucket_name)
        key_name = archive_path.name

        if bucket.get_key(archive_path.name) and not force:
            logging.info('Archive already exists on S3. Skipping upload.')
            return

        logging.info('Uploading archive to S3.')

        k = bucket.new_key(key_name)
        file_size = archive_path.stat().st_size

        bar = tqdm(total=file_size, unit='B', unit_scale=True)

        def percent_cb(complete, total):
            if not complete:
                return
            bar.update(complete - bar.n)

        k.set_contents_from_filename(str(archive_path), cb=percent_cb, num_cb=100)
        bar.close()

    def fetch(self, force=False):
        archive_file_name = self.archive_name_for_project()
        archive_path = self.archives_directory_path / archive_file_name
        if archive_path.exists() and not force:
            return archive_path

        logging.info("Downloading archive {} from {}".format(archive_file_name, self.bucket_name))

        connection = boto.connect_s3(self.AWS_ACCESS_KEY_ID, self.AWS_SECRET_ACCESS_KEY)
        bucket = connection.get_bucket(self.bucket_name)

        key = bucket.get_key(archive_file_name)

        if not key:
            raise PunicRepresentableError("No cached archive with key {}. Are you sure you called `punic cache publish`?".format(archive_file_name))

        content_length = int(key.content_length)

        bar = tqdm(total=content_length, unit='B', unit_scale=True)

        def percent_cb(complete, total):
            if not complete:
                return
            bar.update(complete - bar.n)

        temp_dir = tempfile.mkdtemp()
        temp_archive_path = Path(temp_dir) / archive_file_name

        key.get_contents_to_filename(str(temp_archive_path), cb=percent_cb, num_cb=100)
        bar.close()

        if force and archive_path.exists():
            archive_path.unlink()

        copyfile(temp_archive_path, archive_path)

        return archive_path

    def install(self):
        archive_path = self.fetch(force=False)

        temp_dir = Path(tempfile.mkdtemp())

        logging.info('Expanding archive.')
        with zipfile.ZipFile(str(archive_path)) as archive:
            archive.extractall(str(temp_dir))

        if self.config.build_path.exists():
            rmtree(self.config.build_path)

        logging.info('Replacing {}.'.format(self.config.build_path))
        move(temp_dir, self.config.build_path)


def walk_directory(path):
    for root, dirs, files in os.walk(str(path)):
        for file in files:
            yield (Path(root) / file)



from __future__ import division, absolute_import, print_function

import contextlib
import logging

class RepositoryNotClonedError(Exception):
    pass


class CartfileNotFound(Exception):
    def __init__(self, path):
        self.path = path


class PunicRepresentableError(Exception):
    pass


class NoSuchRevision(Exception):
    def __init__(self, repository, revision):
        self.repository = repository
        self.revision = revision
        self.message = "No such revision <rev>'{}'</rev> in repository <ref>'{}'</ref>.".format(self.revision, self.repository)


class XcodeVersionError(Exception):
    pass


@contextlib.contextmanager
def error_handling():
    try:
        yield
    except RepositoryNotClonedError:
        logging.error('<err>Error</err>: No locally cloned repository found. Did you neglect to run `punic fetch` first?')
        exit(-1)
    except CartfileNotFound as e:
        logging.error('<err>Error</err>: No Cartfile found at path: <ref>{}</ref>'.format(e.path))
        exit(-1)
    except NoSuchRevision as e:
        logging.error('<err>Error</err>: No such revision {} found in repository {}'.format(e.revision, e.repository))
        logging.error('Are you sure you are using the latest bits? Try an explicit `punic fetch` or use `punic bootstrap` instead of `punic build`')
        exit(-1)
    except PunicRepresentableError as e:
        logging.error(e.message)
        exit(-1)
    except:
        raise



from __future__ import division, absolute_import, print_function

__all__ = ['version_check']

import requests
import punic
import logging

from .semantic_version import *


def version_check(verbose=False, timeout=0.3, failure_is_an_option=True):
    try:
        log = logging.info if verbose else logging.debug

        log('<sub>Checking punic version...</sub>')

        current_version = SemanticVersion.string(punic.__version__)
        response = requests.get('https://pypi.python.org/pypi/punic/json', timeout=timeout)
        version_string = response.json()["info"]["version"]
        latest_version = SemanticVersion.string(version_string.strip())

        log('Current version: <rev>{}</rev>, latest version: <rev>{}</rev>'.format(current_version, latest_version))
        if current_version < latest_version:
            logging.warning("""You are using version <rev>{}</rev>, version <rev>{}</rev> is available.
            Use <echo>`pip install -U punic`</echo> to update to latest version.""".format(current_version, latest_version))
    # TODO: Duplicated code
    except requests.exceptions.ReadTimeout as e:
        logging.debug('<error>Failed to check versions')
        if not failure_is_an_option:
            raise e
    except requests.exceptions.ConnectTimeout as e:
        logging.debug('<error>Failed to check versions')
        if not failure_is_an_option:
            raise e



from __future__ import division, absolute_import, print_function

__version__ = '0.2.10'
__all__ = ['Punic', 'current_session']

import os
from copy import copy
from pathlib2 import Path
import logging
from collections import defaultdict

from .cartfile import Cartfile
from .checkout import Checkout
from .config import config
from .repository import Repository, Revision
from .resolver import Resolver, Node
from .runner import runner
from .specification import ProjectIdentifier, Specification, VersionPredicate, VersionOperator
from .xcode import XcodeBuildArguments
import punic.shshutil as shutil
from .errors import NoSuchRevision

current_session = None


class Punic(object):
    __slots__ = ['root_path', 'config', 'all_repositories', 'root_project']

    def __init__(self, root_path=None):

        global current_session

        if not current_session:
            current_session = self

        if not root_path:
            root_path = Path.cwd()

        self.config = config

        root_project_identifier = ProjectIdentifier(overrides=None, project_name=self.config.root_path.name)

        self.all_repositories = {root_project_identifier: Repository(punic=self, identifier=root_project_identifier, repo_path=self.config.root_path),}

        self.root_project = self._repository_for_identifier(root_project_identifier)

    def _resolver(self):
        return Resolver(root=Node(self.root_project.identifier, None), dependencies_for_node=self._dependencies_for_node)

    def _dependencies_for_node(self, node):
        assert not node.version or isinstance(node.version, Revision)
        dependencies = self.dependencies_for_project_and_tag(identifier=node.identifier, tag=node.version)
        return dependencies

    def resolve(self):
        # type: (bool)
        resolver = self._resolver()
        build_order = resolver.resolve_build_order()

        for index, value in enumerate(build_order[:-1]):
            dependency, version = value
            logging.debug('{} <ref>{}</ref> <rev>{}</rev> <ref>{}</ref>'.format(index + 1, dependency, version.revision if version else '', dependency.remote_url))

        specifications = [Specification(identifier=dependency, predicate=VersionPredicate('"{}"'.format(version.revision))) for dependency, version in build_order[:-1]]
        logging.debug("<sub>Saving</sub> <ref>Cartfile.resolved</ref>")

        cartfile = Cartfile(use_ssh=self.config.use_ssh, specifications=specifications)
        cartfile.write((self.config.root_path / 'Cartfile.resolved').open('w'))

    def graph(self):
        # type: (bool) -> DiGraph
        return self._resolver().resolve()

    # TODO: This can be deprecated and the fetch flag relied on instead
    def fetch(self, dependencies=None):

        configuration, platforms = self.config.configuration, self.config.platforms

        if not self.config.build_path.exists():
            self.config.build_path.mkdir(parents=True)

        filtered_dependencies = self._ordered_dependencies(name_filter=dependencies)

        checkouts = [Checkout(punic=self, identifier=identifier, revision=revision) for identifier, revision in filtered_dependencies]
        for checkout in checkouts:
            checkout.prepare()


    def build(self, dependencies):
        # type: ([str])

        logging.info('Using xcode version: {}'.format(self.config.xcode))

        configuration, platforms = self.config.configuration, self.config.platforms

        if not self.config.build_path.exists():
            self.config.build_path.mkdir(parents=True)

        filtered_dependencies = self._ordered_dependencies(name_filter=dependencies)

        checkouts = [Checkout(punic=self, identifier=identifier, revision=revision) for identifier, revision in filtered_dependencies]

        skips = self.config.skips

        def filter_dependency(platform, checkout, project, scheme):
            platform = platform.name
            checkout = checkout.identifier.project_name
            project = project.path.name
            scheme = scheme.name

            for skip in skips:
                current = [ platform, checkout, project, scheme ][:len(skip)]
                if skip == current:
                    # print(skip)
                    # print(current)
                    # print('SKIP?')
                    return False
            return True

        for platform in platforms:
            for checkout in checkouts:
                checkout.prepare()
                for project in checkout.projects:
                    schemes = project.schemes

                    schemes = [scheme for scheme in schemes if scheme.framework_targets]
                    schemes = [scheme for scheme in schemes if platform.device_sdk in scheme.supported_platform_names]
                    for scheme in schemes:
                        if not filter_dependency(platform, checkout, project, scheme):
                            logging.warn('<err>Warning:</err> <sub>Skipping</sub>: {} / {} / {} / {}'.format(platform, checkout.identifier.project_name, project.path.name, scheme.name))
                            continue
                        self._build_one(platform, project, scheme.name, configuration)

    def _ordered_dependencies(self, name_filter=None):
        # type: (bool, [str]) -> [(ProjectIdentifier, Revision)]

        cartfile = Cartfile(use_ssh=self.config.use_ssh, overrides=config.repo_overrides)
        cartfile.read(self.config.root_path / 'Cartfile.resolved')

        def _predicate_to_revision(spec):
            repository = self._repository_for_identifier(spec.identifier)
            if spec.predicate.operator == VersionOperator.commitish:
                try:
                    revision = Revision(repository=repository, revision=spec.predicate.value, revision_type=Revision.Type.commitish, check = True)
                except Exception as e:
                    logging.warning(e.message)
                    return None
                else:
                    return revision
            else:
                raise Exception("Cannot convert spec to revision: {}".format(spec))

        dependencies = [(spec.identifier, _predicate_to_revision(spec)) for spec in cartfile.specifications]
        resolved_dependencies = self._resolver().resolve_versions(dependencies)
        resolved_dependencies = [dependency for dependency in resolved_dependencies if dependency[0].matches(name_filter)]
        return resolved_dependencies

    def _repository_for_identifier(self, identifier):
        # type: (ProjectIdentifier) -> Repository
        if identifier in self.all_repositories:
            return self.all_repositories[identifier]
        else:
            repository = Repository(self, identifier=identifier)
            if self.config.fetch:
                repository.fetch()
            self.all_repositories[identifier] = repository
            return repository

    def dependencies_for_project_and_tag(self, identifier, tag):
        # type: (ProjectIdentifier, Revision) -> [ProjectIdentifier, [Revision]]

        assert isinstance(identifier, ProjectIdentifier)
        assert not tag or isinstance(tag, Revision)

        repository = self._repository_for_identifier(identifier)
        specifications = repository.specifications_for_revision(tag)

        def make(specification):
            repository = self._repository_for_identifier(specification.identifier)
            tags = repository.revisions_for_predicate(specification.predicate)
            if specification.predicate.operator == VersionOperator.commitish:
                try:
                    revision = Revision(repository=repository, revision=specification.predicate.value, revision_type=Revision.Type.commitish, check = True)
                except NoSuchRevision as e:
                    logging.warning("<err>Warning</err>: {}".format(e.message))
                    return None
                tags.append(revision)
                tags.sort()
            assert len(tags)
            return repository.identifier, tags

        dependencies = [make(specification) for specification in specifications]
        dependencies = [dependency for dependency in dependencies if dependency]
        return dependencies

    def _build_one(self, platform, project, scheme, configuration):

        if self.config.dry_run:
            for sdk in platform.sdks:
                logging.warn('<sub>DRY-RUN: (Not) Building</sub>: <ref>{}</ref> (scheme: {}, sdk: {}, configuration: {})...'.format(project.path.name, scheme, sdk, configuration))
            return

        all_products = []

        toolchain = self.config.toolchain

        # Build device & simulator (if sim exists)
        for sdk in platform.sdks:
            logging.info('<sub>Building</sub>: <ref>{}</ref> (scheme: {}, sdk: {}, configuration: {})...'.format(project.path.name, scheme, sdk, configuration))

            derived_data_path = self.config.derived_data_path

            resolved_configuration = configuration if configuration else project.default_configuration
            if not resolved_configuration:
                logging.warn("<err>Warning</err>: No configuration specified for project and no default configuration found. This could be a problem.")

            arguments = XcodeBuildArguments(scheme=scheme, configuration=resolved_configuration, sdk=sdk, toolchain=toolchain, derived_data_path=derived_data_path)

            all_products += project.build(arguments=arguments)

        self._post_process(platform, all_products)

    def _post_process(self, platform, products):
        # type: (punic.platform.Platform, List)

        ########################################################################################################

        logging.debug("<sub>Post processing</sub>...")

        # TODO: QUESTION: Is it possible that this could mix targets with different SDKs?
        products_by_name_then_sdk = defaultdict(dict)
        for product in products:
            products_by_name_then_sdk[product.full_product_name][product.sdk] = product


        for products_by_sdk in products_by_name_then_sdk.values():

            products = products_by_sdk.values()

            # TODO: By convention sdk[0] is always the device sdk (e.g. 'iphoneos' and not 'iphonesimulator')
            primary_sdk = platform.sdks[0]

            device_product = products_by_sdk[primary_sdk]

            ########################################################################################################

            output_product = copy(device_product)
            output_product.target_build_dir = self.config.build_path / platform.output_directory_name

            ########################################################################################################

            logging.debug('<sub>Copying binary</sub>...')
            if output_product.product_path.exists():
                shutil.rmtree(output_product.product_path)

            if not device_product.product_path.exists():
                raise Exception("No product at: {}".format(device_product.product_path))

            shutil.copytree(device_product.product_path, output_product.product_path, symlinks=True)

            ########################################################################################################

            if len(products) > 1:
                logging.debug('<sub>Lipo-ing</sub>...')
                executable_paths = [product.executable_path for product in products]
                command = ['/usr/bin/xcrun', 'lipo', '-create'] + executable_paths + ['-output', output_product.executable_path]
                runner.check_run(command)
                mtime = executable_paths[0].stat().st_mtime
                os.utime(str(output_product.executable_path), (mtime, mtime))

            ########################################################################################################

            logging.debug('<sub>Copying swiftmodule files</sub>...')
            for product in products:
                for path in product.module_paths:
                    relative_path = path.relative_to(product.product_path)
                    shutil.copyfile(path, output_product.product_path / relative_path)

            ########################################################################################################

            logging.debug('<sub>Copying bcsymbolmap files</sub>...')
            for product in products:
                for path in product.bcsymbolmap_paths:
                    shutil.copy(path, output_product.target_build_dir)

            ########################################################################################################

            logging.debug('<sub>Producing dSYM files</sub>...')
            command = ['/usr/bin/xcrun', 'dsymutil', str(output_product.executable_path), '-o', str(output_product.target_build_dir / (output_product.executable_name + '.dSYM'))]
            runner.check_run(command)

            ########################################################################################################


from __future__ import division, absolute_import, print_function

__all__ = ['config_init']

import yaml
from pathlib2 import Path
from prompt_toolkit.contrib.completers import WordCompleter
from prompt_toolkit import prompt
from prompt_toolkit.auto_suggest import (AutoSuggest, Suggestion)
import six
import sys

from .xcode import *
from .platform import *


class ListAutoSuggest(AutoSuggest):
    def __init__(self, items):
        self.items = items

    def get_suggestion(self, cli, buffer, document):

        # Consider only the last line for the suggestion.
        text = document.text.rsplit('\n', 1)[-1]

        # Only create a suggestion when this is not an empty line.
        if text.strip():
            # Find first matching line in history.
            for string in self.items:
                for line in reversed(string.splitlines()):
                    if line.startswith(text):
                        return Suggestion(line[len(text):])


def platform_nicknames():
    return sorted([p.nickname for p in Platform.all])


def _xcode_versions():
    Xcode.find_all()
    return sorted([six.text_type(xcode.version) for xcode in Xcode._all_xcodes])


def _prompt(s, items, default=None):
    items = [six.text_type(item) for item in items]
    completer = WordCompleter(items, ignore_case=True)

    kwargs = {'completer': completer, 'complete_while_typing': True,}

    if default:
        kwargs['default'] = six.text_type(default)

    text = prompt(u'{} ({}): '.format(s, u', '.join(items)), **kwargs)
    if not text:
        return None
    return text


def config_init(**kwargs):
    """Generate punic configuration file."""

    kwargs['xcode_version'] = None

    d = {'defaults': {},}

    configuration = _prompt("Configuration", ['Debug', 'Release'])
    if configuration:
        d['defaults']['configuration'] = configuration

    platform = _prompt("Platform", platform_nicknames())
    if platform:
        d['defaults']['platform'] = platform

    xcode_version = _prompt("Xcode Version", _xcode_versions())
    if xcode_version:
        d['defaults']['xcode-version'] = xcode_version

    stream = six.StringIO()

    yaml.safe_dump(d, stream, default_flow_style=False)

    sys.stdout.write(stream.getvalue())
    if _prompt('Write config to `punic.yaml`', ['yes', 'no'], default='no') == 'yes':
        Path('punic.yaml').open('wb').write(str(stream.getvalue()))



#!/usr/bin/env python

from setuptools import setup

def check_libyaml():
    # TODO: This is a very crude way to detect if libyaml is installed. Should probably move off of libyaml anyways.
    import os
    if not os.path.exists('/usr/local/lib/libyaml.dylib'):
        import sys
        sys.stderr.write('Error: {}\nError: libyaml not installed. Use `brew install libyaml` to install it\nError: {}\n'.format('*' * 80, '*' * 80))
        exit(-1)

check_libyaml()


setup(
    name='punic',
    version='0.2.10',
    url='http://github.com/schwa/punic',
    license='MIT',
    author='Jonathan Wight',
    author_email='jwight@mac.com',
    description='Clean room python implementation of a subset of Carthage functionality',
    packages=['punic'],
    keywords = ['carthage', 'build', 'xcode', 'iOS', 'cocoapods', 'cocoa', 'macOS'],
    platform = "MacOS X",
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: MacOS X',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Build Tools',
        ],
    install_requires=[
        'affirm',
        'blessings',
        'boto',
        'click',
        'click_didyoumean',
        'flufl.enum',
        'jsonpath_rw',
        'memoize',
        'networkx',
        'pathlib2',
        'prompt_toolkit',
        'pyyaml',
        'requests',
        'six',
        'tqdm',
        ],
    entry_points='''
        [console_scripts]
        punic=punic.punic_cli:main
        ''',
)



from __future__ import division, absolute_import, print_function

__all__ = ['styled', 'styled_print']

# noinspection PyUnresolvedReferences
from six.moves.html_parser import HTMLParser
from blessings import Terminal
import six
import sys

term = Terminal()

default_styles = {
    'err': term.red,

    'ref': term.yellow,
    'path': term.yellow,

    'rev': term.bold,
    'version': term.bold,

    'cmd': term.cyan + term.underline,  # 'sub': term.cyan,

    'echo': term.yellow,
}


class MyHTMLParser(HTMLParser):
    def __init__(self, style, styles = None):
        HTMLParser.__init__(self)

        self.s = ''
        self.style = style

        self.styles = styles if styles else default_styles
        self.style_stack = []

    # noinspection PyUnusedLocal
    def handle_starttag(self, tag, attrs):
        if tag in self.styles:
            self.style_stack.append(self.styles[tag])

    def handle_endtag(self, tag):
        if tag in self.styles:
            self.style_stack.pop()

    def handle_data(self, data):
        if self.style:
            self.apply()
        self.s += data

    def apply(self):
        self.s += term.normal
        for style in set(self.style_stack):
            self.s += style

from punic.config import config

def styled(s, style = None, styles = None):

    if style is None:
        style = config.color
    else:
        style = True

    parser = MyHTMLParser(style=style, styles = styles)
    parser.feed(s)
    return parser.s + (term.normal if style else '')



def styled_print(message, sep=' ', end='\n', file=sys.stdout, flush=False, style = None, styles = None, *args):
    args = [message] + list(args)
    s = sep.join([six.text_type(arg) for arg in args]) + end
    s = styled(s, style = style, styles = styles)

    file.write(s)
    if flush:
        file.flush()


# '<head>***</head> Checkout out <title>SwiftLogging</title> at "<version>v1.0.1</version>"')
#
# # instantiate the parser and fed it some HTML



from __future__ import division, absolute_import, print_function

__all__ = ['HTMLFormatter', 'HTMLStripperFormatter']

import punic.styling
import six

# TODO: Convert < and > to &lt; and &gt;

class HTMLFormatter(object):

    def __init__(self):
        self.color = True

    def format(self, record):
        message = record.msg
        if not isinstance(message, six.string_types):
            message = repr(message)
        else:
            message = punic.styling.styled(message, self.color)
        return message


class HTMLStripperFormatter(object):
    def __init__(self, formatter=None):
        self.formatter = formatter

    def format(self, record):
        if self.formatter:
            message = self.formatter.format(record)
        else:
            message = record.msg
        if not isinstance(message, six.string_types):
            message = repr(message)
        else:
            message = punic.styling.styled(message, False)
        return message
